%{
(* Author:  Robbert van Renesse, Cornell University.
 *
 * Parser for .mli files.
 *)

open Camouglob
%}

%token <string> LC_IDENT
%token <string> UC_IDENT
%token <string> STRING
%token <string * string> INCLUDE
%token OF
%token QUOTE
%token LB
%token RB
%token LCB
%token RCB
%token COLON
%token SEMICOLON
%token TIMES
%token COMMA
%token DOT
%token EQUALS
%token OR
%token ARROW
%token MUTABLE
%token TYPE
%token VAL
%token EXTERNAL
%token EXTTYPE
%token EXCEPTION
%token EOF

%start interface
%type <Camouglob.decl list> interface

%%

interface
	: decl_list EOF
		{ $1 }

decl_list
	: declaration
		{ [$1] }
	| declaration decl_list
		{ $1 :: $2 }

declaration
	: TYPE type_name EQUALS type_expr
		{ TYPE_DECL($2, $4) }
	| TYPE type_name
		{ TYPE_DECL($2, ABSTRACT_TYPE) }
	| VAL LC_IDENT COLON type_expr
		{ VAL_DECL($2, $4) }
	| EXTTYPE LC_IDENT
		{ EXTTYPE_DECL($2, $2, ABSTRACT_TYPE) }
	| EXTTYPE LC_IDENT COLON STRING
		{ EXTTYPE_DECL($2, $4, ABSTRACT_TYPE) }
	| EXTTYPE LC_IDENT EQUALS type_expr
		{ EXTTYPE_DECL($2, $2, $4) }
	| EXTTYPE LC_IDENT EQUALS type_expr COLON STRING
		{ EXTTYPE_DECL($2, $6, $4) }
	| EXTERNAL LC_IDENT COLON type_expr
		{ EXTVAL_DECL($2, $2, $4) }
	| EXTERNAL LC_IDENT COLON type_expr EQUALS STRING
		{ EXTVAL_DECL($2, $6, $4) }
	| EXCEPTION type_constr_list
		{ EXCEPTION_DECL($2) }
	| INCLUDE
		{ INCLUDE_DECL $1 }

type_name
	: LC_IDENT
		{ ($1, []) }
	| alpha_type LC_IDENT
		{ ($2, [$1]) }
	| LB alpha_type_list RB LC_IDENT
		{ ($4, $2) }

alpha_type
	: QUOTE LC_IDENT
		{ $2 }

alpha_type_list
	: alpha_type
		{ [$1] }
	| alpha_type COMMA alpha_type_list
		{ $1 :: $3 }

type_expr
	: type_arg
		{ $1 }
	| type_constr_list
		{ CONSTR_TYPE($1) }
	| OR type_constr_list
		{ CONSTR_TYPE($2) }

type_arg
	: type_product_list
		{ match $1 with
		    | [] -> failwith "impossible"
		    | [singleton] -> singleton
		    | hd :: tl -> FUNCTION_TYPE(List.rev tl, hd)
		}

type_product_list
	: type_product
		{ [$1] }
	| type_product_list ARROW type_product
		{ $3 :: $1 }

type_product
	: type_term_list
		{ match $1 with
		    | [] -> failwith "impossible"
		    | [singleton] -> singleton
		    | hd :: tl as l -> PRODUCT_TYPE l
		}

type_term_list
	: type_term
		{ [$1] }
	| type_term TIMES type_term_list
		{ $1 :: $3 }

type_term
	: lc_ident
		{ NAMED_TYPE($1, []) }
	| type_term lc_ident
		{ NAMED_TYPE($2, [$1]) }
	| LB type_expr COMMA type_expr_list RB lc_ident
		{ NAMED_TYPE($6, ($2 :: $4)) }
	| alpha_type
		{ ALPHA_TYPE($1) }
	| LCB type_decl_list RCB
		{ STRUCT_TYPE($2) }
	| LB type_expr RB
		{ $2 }

lc_ident
	: LC_IDENT
		{ ("", $1) }
	| UC_IDENT DOT LC_IDENT
		{ ($1, $3) }

type_constr_list
	: type_constr
		{ [$1] }
	| type_constr OR type_constr_list
		{ $1 :: $3 }

type_constr
	: UC_IDENT
		{ CCONST $1 }
	| UC_IDENT OF type_arg
		{ NCCONST($1, $3) }

type_expr_list
	: type_expr
		{ [$1] }
	| type_expr COMMA type_expr_list
		{ $1 :: $3 }

type_decl
	: LC_IDENT COLON type_expr
		{ ($1, false, $3) }
	| MUTABLE LC_IDENT COLON type_expr
		{ ($2, true, $4) }

type_decl_list
	: type_decl
		{ [$1] }
	| type_decl SEMICOLON type_decl_list
		{ $1 :: $3 }
